#!/bin/sh
: "This is a shell archive, meaning:                              "
: "1. Remove everything above the #! /bin/sh line.                "
: "2. Save the resulting test in a file.                          "
: "3. Execute the file with /bin/sh (not csh) to create the files:"
: "	pcal.c"
: "	pcal.man"
: "	pcal.ps"
: "This archive created:  Tue Apr 17 14:01:25 MET DST 1990 "
echo file: pcal.c
sed 's/^X//' >pcal.c << 'END-of-pcal.c'
X/*
X * pcal.c: PostScript/C monthly calendar, with items from the user's
X *		calendar file placed accordingly.
X *
X * Revision History:
X *
X *		kk: Ken Keirnan of Pacific Bell
X *		pw: Patrick Wood of Pipeline Associates
X *		bv: Bill Vogel of AT&T.
X *		jpl: Joe Larson of Dayton-Hudson
X *
X *		?-87	pw		PostScript code written.
X *		?-??	bv		Calendar file inclusion code written.
X *		?-89	kk		Collected together and posted to Usenet.
X *		1-90	jpl.	Fixed bug in postscript code.  Added a full-year
X *						mode.  Posted to Usenet.
X *		1-90	jpl.	Implemented optional rotation, fonts, etc as per
X *						suggestions (and even some "diffs") from the net.
X *						Split prolog into external file.
X *						Fixed the "-m month" option so it really works.
X *						Got rid of some questionable C practices due to
X *						reports of syntax errors with some compilers.
X */
X
X#include <stdio.h>
X#include <ctype.h>
X#include <time.h>
X#include <string.h>
X
X#ifndef	PROLOG_FILE			/* Define this file accordingly. */
X#define	PROLOG_FILE	"/RIA/vega/users/matthieu/PostScript/pcal/pcal.ps"
X#endif
X
Xchar	*words[100],		/* maximum number of words on a line */
X		lbuf[512];			/* maximum line size */
X
Xchar months[12][4] =		/* used to match alpha months */
X{
X	"jan", "feb", "mar", "apr", "may", "jun",
X	"jul", "aug", "sep", "oct", "nov", "dec"
X};
X
X
XFILE	*cfp = NULL;		/* Calendar input file */
Xint		year,
X		cyear;
X
Xextern int		errno;
Xvoid exit();
Xchar *getenv();
X
Xmain(argc, argv)
Xint argc;
Xchar **argv;
X{
X	extern char		*optarg;
X	struct tm		*lt;
X	FILE			*ifile;				/* Prolog input file		*/
X
X	char			cbuf[80],			/* String buffer			*/
X					*cfile,				/* Name of calendar file	*/
X					*cp,				/* Used in the getenv stuff	*/
X 					*titlefont,
X					*dayfont;
X
X	long			t, time();
X 	short			rotate,
X					nocal,
X					m,
X					month;
X
X	char			doyear,
X					doprolog;
X
X
X/*
X * Initialize variables.
X */
X	cfile = NULL;
X	titlefont = "Times-Bold";
X	dayfont = "Times-Bold";
X	rotate = 90;
X	nocal = month = 0;
X	doyear = 0;
X	doprolog = 1;
X
X/*
X * Handle the arglist.
X */
X 	while ((m = getopt(argc, argv, "ef:pr:m:y:t:d:")) != EOF)
X		switch (m)
X		{
X			case 'p':	doprolog = 0; break;
X			case 'e':	nocal++; cfile = NULL; break;
X
X			case 'f':	cfile = optarg; nocal = 0; break;
X
X			case 'm':	month = atoi(optarg); if (!month) doyear = 1; break;
X
X			case 'y':	year = atoi(optarg);
X						if (year < 1900) year = year % 100 + 1900;
X						break;
X
X			case 'r':	rotate = 0; break;
X			case 't':	titlefont = optarg; break;
X			case 'd':	dayfont = optarg; break;
X
X			case '?':	fprintf(stderr, "\
XUsage: %s [ -r ] [ -e | -f <cal> ] [ -m month] [ -y <year> ]\n\
X        [ -t <title font> ] [ -d <day font> ]\n", argv[0]);
X						exit(1);
X		}
X
X
X/*
X * Get all the month/year info needed.
X */
X	t = time((long *)0);
X	lt = localtime(&t);
X
X	if (!month && !doyear)
X		month = lt->tm_mon + 1;
X	if (!year)
X		year = lt->tm_year + 1900;
X	cyear = year;		/* If no year in calendar file, assume current. */
X
X/*
X * Open a supplied calendar file (if any)
X */
X	if (cfile != NULL)
X	{
X		if ((cfp = fopen(cfile, "r")) == NULL)
X		{
X			fprintf(stderr, "pcal: can't open file: %s\n", cfile);
X			exit(1);
X		}
X	}
X/*
X * Else see if a calendar file exists in the home directory
X */
X	else if (nocal == 0 && (cp = getenv("HOME")) != NULL)
X	{
X		(void)strcpy(cbuf, cp);
X		(void)strcat(cbuf, "/calendar");
X		cfp = fopen(cbuf, "r");
X	}
X
X/*
X * Write out PostScript prolog
X */
X	if (doprolog)
X	{
X		if ((ifile = fopen(PROLOG_FILE, "r")) == NULL)
X		{
X			perror(PROLOG_FILE);
X			exit(errno);
X		}
X		printf("%%!\n");
X
X		while (fgets(cbuf, 80, ifile) > 0)
X			printf("%s", cbuf);
X		fclose(ifile);
X	}
X
X	printf("/titlefont /%s def\n/dayfont /%s def\n", titlefont, dayfont);
X 	printf("%d rotate\n", rotate);
X 	if (rotate)
X 		printf("50 -120 translate\n");
X 	else
X 		printf("0.75 0.75 scale\n50 460 translate\n");
X
X/*
X * Now go do the monthly calendar(s).
X */
X	printf("/year %d def\n", year);
X	if (month)
X		pmonth(month);
X	else
X		for (month = 1; month <= 12; month++)
X			pmonth(month);
X
X	return(0);
X}
X
X/*
X * pmonth - do calendar for month "m"
X */
Xpmonth(m)
Xint m;
X{
X	register char **s;
X	register oldday;
X	register day;
X
X	oldday = -1;
X	/*
X	 * Do the calendar
X	 */
X	printf("/month %d def\n", m);
X	printf("printmonth\n");
X
X	/*
X	 * Browse through the calendar file looking for day info
X	 */
X	if ((day = getday(m)) == 0)		/* no info */
X	{
X		printf("showpage\n");
X		return;
X	}
X	do
X	{
X		if (day != oldday)
X		{
X			if (oldday == -1)
X				printf("%d [ \n", day);
X			else
X				printf(" ] daytext %d [  \n", day);
X			oldday = day;
X		}
X		else
X			printf("(.p)\n");
X		for (s = words; *s; s++)
X			printf("(%s)\n", *s);
X
X	} while (day = getday(m));
X	printf("] daytext\n");
X	printf("showpage\n");
X}
X
X/*
X * getday - find next day entry for desired month in the calendar file
X */
Xgetday(m)
Xregister m;
X{
X	static mon = 0;
X	static eof = 0;
X	register char *cp;
X	register c;
X
X	if (cfp == NULL)	/* whoops, no calendar file */
X		return(0);
X
X	if (m != mon)		/* new month, rewind */
X	{
X		rewind(cfp);
X		eof = 0;
X		mon = m;
X	}
X	if (eof)
X		return(0);
X	while (1)
X	{
X		cp = lbuf;
X		do
X		{
X			while ((c = getc(cfp)) != '\n' && c != EOF)
X			{
X				/* ignore leading white space */
X				if (cp == lbuf && (c == ' ' || c == '\t'))
X					continue;
X				*cp++ = c;
X			}
X			if (c == EOF)
X			{
X				eof = 1;
X				return(0);
X			}
X		} while (cp == lbuf);	/* ignore empty lines */
X		*cp = 0;
X
X		/* examine the line, see if its one we want */
X		if ((c = parse(m)) != 0)
X			break;
X	}
X
X	return(c);
X}
X
X/*
X * parse - check calendar entry for desired month, break line into fields
X */
Xparse(m)
Xregister m;
X{
X	register char *cp;
X	register i;
X
X	cp = strtok(lbuf, " \t");	/* get first field */
X	while (*cp)
X	{
X		if (isupper(*cp))
X			*cp = tolower(*cp);
X		cp++;
X	}
X	cp = lbuf;
X
X	/*
X	 * Check for "year" line
X	 */
X	if (strcmp(cp, "year") == 0)
X	{
X		cp = strtok((char *)0, " \t");
X		if ((i = atoi(cp)) > 0)
X		{
X			if (i < 100)
X				i += 1900;
X			cyear = i;
X		}
X		return(0);
X	}
X	/*
X	 * If field begins with alpha, try to decode month name
X	 */
X	if (isalpha(*cp))
X	{
X		if (cyear != year)
X			return(0);
X
X		for (i = 0; i < 12; i++)
X			if (strncmp(cp, months[i], 3) == 0)
X			{
X				if (++i != m)
X					return(0);
X
X				/* month found, get day */
X
X				if ((cp = strtok((char *)0, " \t")) == NULL)
X					return(0);
X				if ((i = atoi(cp)) < 1 || i > 31)
X					return(0);
X				if (loadwords())
X					return(i);
X				return(0);
X			}
X		return(0);
X	}
X	/*
X	 * Not alpha month, try numeric
X	 */
X	if ((i = atoi(cp)) != m)
X		return(0);
X	while (isdigit(*cp))
X		cp++;
X	while (*cp && !isdigit(*cp))
X		cp++;
X
X	/* now get day */
X
X	if ((i = atoi(cp)) < 1 || i > 31)
X		return(0);
X
X	 /* Numeric dates may have a year */
X
X	while (isdigit(*cp))
X		cp++;
X	while (*cp && !isdigit(*cp))
X		cp++;
X	if ((m = atoi(cp)) > 0)
X	{
X		if (m < 100)
X			m += 1900;
X		cyear = m;
X	}
X
X	if (cyear != year)
X		return(0);
X
X	if (loadwords())
X		return(i);
X	return(0);
X}
X
X/*
X * loadwords - tokenize line buffer into word array
X */
Xloadwords()
X{
X	register char **ap = words, *c;
X	register i;
X
X	for (i = 0; *ap = strtok((char *)0, " \t") ; ap++, i++);
X	return(i);
X}
END-of-pcal.c
echo file: pcal.man
sed 's/^X//' >pcal.man << 'END-of-pcal.man'
X.TH PCAL 1
X.SH NAME
Xpcal \- generate PostScript calendars
X.SH SYNOPSIS
X.B pcal
X[
X.B \-e
X|
X.BR \-f <cal>
X] [
X.BR \-m <month>
X] [
X.BR \-y <year>
X] [
X.B \-r
X] [
X.BR \-t <title-font>
X] [
X.BR \-d <day-font>
X] [
X.B \-p
X]
X.SH DESCRIPTION
X.I Pcal
Xgenerates PostScript to produce landscape calendars for any month and year.
XThe arguments
X.B <month>
Xand
X.BR <year> ,
Xif provided, should be numeric.  The month should be in the range 1 - 12,
Xand year should be specified as 1 or 2 digits or as the full 4 digit year.
XThe defaults for month and year are the current month and year.
XIf month is provided, but zero, the entire year is printed.
X.P
XIf a file named
X.I calendar
Xresides in the callers home directory, it will be searched for lines with
Xleading dates matching the requested month and year (current by default).
XAny text following the date will be printed on the calendar under the
Xappropriate day of the month.  Dates in the
X.I calendar
Xfile may consist of a numeric or alpha month (at least the first 3 characters
Xfor month names) followed by a numeric day and optionally followed by a
Xyear.  Any non-numeric character may separate numeric dates.  Lines in the
X.I calendar
Xfile consisting of "year xxxx" (where xxxx is a numeric year) can be used
Xto set the year for following entries.  This assumes that the following
Xentries do not contain a year.  Any date entries containing year information
Xwill set the remembered year to that year.
X.P
X.I Pcal has options:
X.P
X.TP
X.B \-e
XPrint an empty calendar.  Do not print entries from a calendar file.
X.TP
X.BR \-f <cal>
XDirects
X.I pcal
Xto use the file name <cal> as the input file in place of the default
Xcalendar file in the callers home directory.
X.TP
X.BR \-t <title-font>
XSelects the PostScript font name for titles.
X.TP
X.BR \-d <day-font>
XSelects the PostScript font name for day names.
X.TP
X.B \-r
XDo not rotate the calendar (ie: print portrait rather than landscape).
X.TP
X.B \-p
XDoes not print the PostScript prolog code.
X.SH SEE ALSO
Xcal(1)
END-of-pcal.man
echo file: pcal.ps
sed 's/^X//' >pcal.ps << 'END-of-pcal.ps'
X%!
X% /titlefont /Times-Bold def
X% /dayfont /Helvetica-Bold def
X/month_names [ (January) (February) (March) (April) (May) (June) (July)
X		(August) (September) (October) (November) (December) ] def
X/prtnum { 3 string cvs show} def
X/drawgrid 		% draw calendar boxes
X{
X	dayfont findfont 10 scalefont setfont
X	0 1 6
X	{
X		dup dup 100 mul 40 moveto
X		[ (Sunday) (Monday) (Tuesday) (Wednesday) (Thursday) (Friday) (Saturday) ] exch get
X		100 center
X		100 mul 35 moveto
X		1.0 setlinewidth
X		0 1 5
X		{
X			gsave
X			100 0 rlineto 
X			0 -80 rlineto
X			-100 0 rlineto
X			closepath stroke
X			grestore
X			0 -80 rmoveto
X			pop
X		 } for
X	} for
X} def
X
X
X/drawnums		% place day numbers on calendar
X{
X	dayfont findfont 30 scalefont setfont
X	/start startday def
X	/days ndays def
X	start 100 mul 5 add 10 rmoveto
X	1 1 days
X	{
X		/day exch def
X		gsave
X%  #ifndef SATBLK
X%  		day start add 7 mod 0 eq
X%  		{
X%  			submonth 0 eq
X%  			{
X%  				.8 setgray
X%  			} if
X%  		} if
X%  #endif
X		day start add 7 mod 1 eq
X		{
X			submonth 0 eq
X			{
X				.8 setgray
X			} if
X		} if
X		day prtnum
X		grestore
X		day start add 7 mod 0 eq
X		{
X			currentpoint exch pop 80 sub 5 exch moveto
X		}
X		{
X			100 0 rmoveto
X		} ifelse
X	} for
X} def
X
X/drawfill 		% place fill squares on calendar
X{
X	/start startday def
X	/days ndays def
X	0 35 rmoveto
X	1.0 setlinewidth
X	0 1 start 1 sub
X	{
X		gsave
X		.9 setgray
X		100 0 rlineto 
X		0 -80 rlineto
X		-100 0 rlineto
X		closepath fill
X		grestore
X		100 0 rmoveto
X		pop
X	} for
X	submonth 1 eq
X	{
X		/lastday 42 def
X		600 -365 moveto
X	}
X	{
X		/lastday 40 def
X		400 -365 moveto
X	} ifelse
X	lastday -1 ndays start 1 add add
X	{
X		/day exch def
X		gsave
X		.9 setgray
X		100 0 rlineto 
X		0 -80 rlineto
X		-100 0 rlineto
X		closepath fill
X		grestore
X		day 7 mod 1 eq
X		{
X			600 -365 80 add moveto
X		}
X		{
X			-100 0 rmoveto
X		} ifelse
X	} for
X} def
X
X/isleap		% is this a leap year?
X{
X	year 4 mod 0 eq		% multiple of 4
X	year 100 mod 0 ne 	% not century
X	year 1000 mod 0 eq or and	% unless it's a millenia
X} def
X
X/days_month [ 31 28 31 30 31 30 31 31 30 31 30 31 ] def
X/ndays		% number of days in this month
X{
X	days_month month 1 sub get
X	month 2 eq	% Feb
X	isleap and
X	{
X		1 add
X	} if
X} def
X
X/startday		% starting day-of-week for this month
X{
X	/off year 2000 sub def	% offset from start of epoch
X	off
X	off 4 idiv add		% number of leap years
X	off 100 idiv sub	% number of centuries
X	off 1000 idiv add	% number of millenia
X	6 add 7 mod 7 add 	% offset from Jan 1 2000
X	/off exch def
X	1 1 month 1 sub
X	{
X		/idx exch def
X		days_month idx 1 sub get
X		idx 2 eq
X		isleap and
X		{
X			1 add
X		} if
X		/off exch off add def
X	} for
X	off 7 mod		% 0--Sunday, 1--monday, etc.
X} def
X
X/center		% center string in given width
X{
X	/width exch def
X	/str exch def width str 
X	stringwidth pop sub 2 div 0 rmoveto str show
X} def
X
X/calendar
X{
X	titlefont findfont 48 scalefont setfont
X	0 60 moveto
X	/month_name month_names month 1 sub get def
X	month_name show
X	/yearstring year 10 string cvs def
X	700 yearstring stringwidth pop sub 60 moveto
X	yearstring show
X	0 0 moveto
X	drawnums
X	0 0 moveto
X	drawfill
X	0 0 moveto
X	drawgrid
X} def
X
X/daytext
X{
X	/Helvetica-Narrow findfont 6 scalefont setfont
X	/mytext	exch def /myday exch def
X	startday myday 1 sub add dup 7 mod 100 mul 5 add % gives column
X	exch 7 idiv -80 mul % gives row
X	dup /ypos exch def moveto
X	/LM currentpoint pop def /RM LM 95 add def
X    mytext { dup (.p) eq { crlf pop} {prstr ( ) show} ifelse } forall
X} def
X
X/crlf
X{
X	ypos 8 sub /ypos exch def LM ypos moveto
X} def
X
X/prstr
X{
X	dup stringwidth pop currentpoint pop
X	add RM gt {crlf} if show
X} def
X
X/printmonth
X{
X%	90 rotate
X%	50 -120 translate
X	/submonth 0 def
X	calendar
X	month 1 sub 0 eq
X	{
X		/lmonth 12 def
X		/lyear year 1 sub def
X	}
X	{
X		/lmonth month 1 sub def
X		/lyear year def
X	} ifelse
X	month 1 add 13 eq
X	{
X		/nmonth 1 def
X		/nyear year 1 add def
X	} 
X	{
X		/nmonth month 1 add def
X		/nyear year def
X	} ifelse
X	/savemonth month def
X	/saveyear year def
X	/submonth 1 def
X	/year lyear def
X	/month lmonth def
X	gsave
X	500 -365 translate
X	gsave
X	.138 .138 scale
X	10 -120 translate
X	calendar
X	grestore
X	/submonth 1 def
X	/year nyear def
X	/month nmonth def
X	100 0 translate
X	gsave
X	.138 .138 scale
X	10 -120 translate
X	calendar
X	grestore
X	/month savemonth def
X	/year saveyear def
X	/submonth 0 def
X	grestore
X} def
END-of-pcal.ps
exit

------- End of Forwarded Message

