#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>

#include "Calendar.h"
#include "structs.h"
#include "MuMenu.h"

#include <Xm/Xm.h>
#include <Xm/Mu.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/DrawingA.h>
#include <Xm/ToggleB.h>

char filename[50];

#define MAXMONTHS 100
#define W_SUCCESS 0
#define W_BADENTRY 1
#define W_FILE 2

#define MAXMINS 6

int mins[MAXMINS] = { 120, 30, 10, 5, 2, 1 };

CalendarSelection *(selections[MAXMONTHS]);
int nummonths;

extern Widget top[2];
extern char *suffix();
extern char *month();

Boolean editingold;
desc *editingwhat;

Display *disp;
Window root;
Screen *screen;
int screennum;

Widget mainShell;
Widget mainForm;

Widget new, add, descr, stime, etime, from, to;
Widget header, dayList, cal1, cal2, menubar;
Widget fwidget[3];
Widget parmForm, schedToggle, minHead, minToggle[6];

CalendarWidget displayedcals[13];

Arg args[10];
CalendarSelection displayed;

int numitems = 0;
simpleEvent *itemarray[50];

event *events;
simpleEvent *simp;
int basemonth, baseyear;

void alldone();
void saveproc();
void printproc();
void newMonth();
void newYear();
void togpref();
void help();
void deleteEnt();
void showEdit();
extern void createApptWidgets();
extern void appts();
extern void writepeople();
extern void flush();
extern void hide();
extern void apptHelp();
extern void deletef();
int writerec();

static MuCallbacksRec callbacks[] = {
  {"quit",           alldone},
  {"save",           saveproc},
  {"print",          printproc},
  {"showEdit",       showEdit},
  {"deleteEnt",      deleteEnt},
  {"month",          newMonth},
  {"year",           newYear},
  {"togpref",        togpref},
  {"appts",          appts},
  {"help",           help},
  {"flush",          flush},
  {"hide",           hide},
  {"apptHelp",       apptHelp},
  {"deletef",        deletef}
};

void alldone( widget, closure, call_data )
Widget widget;
caddr_t closure;
caddr_t call_data;
{
  writepeople();
  exit();
}

char *helptext[] = {
"   This is a calendar program currently under development, with a bizillion\n\
features to come. The current user interface is a substitute for the real\n\
interface, which hasn't been written yet. The file format it uses will at\n\
least be upward compatible, so it is safe to start using the program now\n\
if you find it to be useful.\n\n\
   Your appointments are stored in a file called .tc in your home directory.\n\
Your userlist (in the appointments window) is stored in a file called\n\
.people. They will eventually occupy a subdirectory.",
"   When you select a single day on one of the calendars (done by clicking\n\
in the day), the appointments entered for that day will be displayed in a\n\
list below the calendar. Clicking on an item in that list will in turn\n\
display the full description for that item in boxes below where the\n\
information may be edited.\n\n\
   Once the information has been changed, just click the enter button and\n\
the new information will be reentered into the list. You may also extend\n\
the selection of dates a particular event occupies. Just use shift-click\n\
in the calendar to add or remove additional dates. If you shift-click a\n\
selected date, it will be deselected. Dragging through dates in the\n\
calendar also works. Experiment. An item may have as many dates\n\
associated with it in as many months as you like.\n\n\
   There are scroll bars now, but they're buggy. I will fix that soon.\n\
It should only core dump, or go into an infinite loop,  when you resize\n\
the window, and then only sometimes.",
"   The editing areas have emacs bindings set up on them. Tabbing from one\n\
of the time edit boxes will move focus to the next box; shift-tab will move\n\
you out of the description editing box. Meta-return in any of the editing\n\
boxes is equivalent to clicking on enter.\n\n\
   There is a preferences option which affects the interpretation of times\n\
you enter. Assume daytime assumes any hour you enter from 8 to 7 that\n\
doesn't have an AM or PM on the end to be in the range 8AM-7PM. If assume\n\
daytime is not selected, then 24-hour format is assumed for times without\n\
AM/PM. (AM/PM is not case sensitive; only A/P are required.)\n\n\
   The edit menu is totally nonfunctional.",
"   You can pop up and hide the appointments window by selecting the\n\
appointments option from the Windows menu. There is more help about\n\
appointments in the help menu of that window.",
"   To delete an entry, select it first and then select \"Delete selected\n\
entry\".",
"   The file is currently saved every time you make an entry. The save\n\
option on the quit menu is therefore usually redundant, which is good\n\
because it isn't fully implemented yet.",
"   The print option under the file menu asks for a printer name (defaulting\n\
to nil). Once you click OK, it prints the currently displayed month to\n\
that printer. The \"currently displayed month\" is the leftmost of the\n\
two months that are displayed. Just use the Month and Year menus to change\n\
it as usual.\n\
\n\
   The fancy postscript stuff is code stolen from Usenet.",
"   There are currently many bugs implemented. Of course, you don't know\n\
which once are implemented and which ones are accidental, so go ahead and\n\
report them all. The reason for implemeted bugs is that to have done the\n\
code right the first time (or at all) was pointless, since a lot of it's\n\
just going to get thrown out again soon anyway. The most important bugs\n\
I want to know about are those that cause data loss or corruption. I don't\n\
know of any that exist, and will promptly fix any reported ones that are\n\
reproducible or obvious (i.e., the ones that I can figure out).\n\n\
   Send bug reports to cfields. Thanks!"
};

void help( widget, closure, call_data )
Widget widget;
int closure;
caddr_t call_data;
{
  MyHelp( 0, helptext[closure-1] );
}

clearSelRecs()
{
  nummonths = 0;
}

int getAct()
{
  int i, bits = 0;
  Boolean bit;
  Arg args;

  for ( i = 0; i < MAXMINS; i++ )
    {
      XtSetArg( args, XmNset, &bit );
      XtGetValues( minToggle[i], &args, 1 );
      bits *= 2;
      if ( bit )
	bits++;
    }
  return bits;
}

void setAct( bits )
int bits;
{
  int i;
  Boolean bit;
  Arg args;
  
  for ( i = MAXMINS - 1; i >= 0; i-- )
    {
      bit = bits&1;
      bits /= 2;
      XtSetArg( args, XmNset, bit );
      XtSetValues( minToggle[i], &args, 1 );
    }
}

setSelRecs( sel )
CalendarSelection sel[];
{
  int i, j, found = 0;

  for ( i = 0; i < nummonths && !found; i++ )
    {
      if ( (selections[i])[0].month == sel[0].month &&
	   (selections[i])[0].year == sel[0].year )
	{
	  found = 1;
	  j = 0;
	  while ( sel[j].day != 0 )
	    {
	      (selections[i])[j] = sel[j];
	      j++;
	    }
	  (selections[i])[j] = sel[j];
/*
	  (selections[i])[j].
	  (selections[i])[j].day = 0;
	  (selections[i])[j].year = sel[j]
*/
	}
    }
  if ( !found )
    {
      nummonths++; /* bug */
      if ( selections[nummonths-1] == NULL )
	selections[nummonths-1] = (CalendarSelection *)
	  XtMalloc( 32 * sizeof( CalendarSelection ) );

      i = nummonths - 1;
      j = 0;

      (selections[i])[0].month = sel[0].month; /* in case sel is empty */
      (selections[i])[0].year = sel[0].year;

      while ( sel[j].day != 0 )
	{
	  (selections[i])[j] = sel[j];
	  j++;
	}
      (selections[i])[j] = sel[j];
/*      (selections[i])[j].day = 0; */
    }
}

getSelRecs( sel )
CalendarSelection sel[];
{
  int i, j, found = 0;

  for ( i = 0; i < nummonths && !found; i++ )
    {
      if ( (selections[i])[0].month == sel[0].month &&
	   (selections[i])[0].year == sel[0].year )
	{
	  found = 1;
	  j = 0;
	  while ( (selections[i])[j].day != 0 )
	    {
	      sel[j] = (selections[i])[j];
	      j++;
	    }
	  sel[j] = (selections[i])[j];
/*	  sel[j].day = 0; */
	}
    }
  if ( !found )
    sel[0].day = 0; /* rest is already set */
}

int noselections()
{
  int i;

  for ( i = 0; i < nummonths; i++ )
    if ( (selections[i])[0].day != 0 )
      return 0;
  return 1;
}

printSelRecs()
{
  int i, j;

  for ( i = 0; i < nummonths; i++ )
    for ( j = 0; (selections[i])[j].day != 0; j++ )
      fprintf( stdout, "%d/%d/%d ",
	      (selections[i])[j].month,
	      (selections[i])[j].day,
	      (selections[i])[j].year );
  fprintf( stdout, "\n" );
}

void newMonth( widget, closure, call_data )
Widget widget;
int closure;
caddr_t call_data;
{
  CalendarSelection sel[32];

  int n, nextmonth, nextyear;

  basemonth = closure;
  nextmonth = basemonth%12 + 1;
  nextyear = baseyear + ((basemonth == 12)?1:0);

  n = 0;
  XtSetArg( args[n], XmNmonth, basemonth ); n++;
  XtSetValues( cal1, args, n );

  sel[0].month = basemonth;
  sel[0].year = baseyear;
  sel[0].day = 0;
  getSelRecs( sel );
  CalSetSelInfo( cal1, SELECTED, sel );

  n = 0;
  XtSetArg( args[n], XmNmonth, nextmonth ); n++;
  XtSetArg( args[n], XmNyear, nextyear ); n++;
  XtSetValues( cal2, args, n );

  sel[0].month = nextmonth;
  sel[0].year = nextyear;
  sel[0].day = 0;
  getSelRecs( sel );
  CalSetSelInfo( cal2, SELECTED, sel );
}

void newYear( widget, closure, call_data )
Widget widget;
int closure;
caddr_t call_data;
{
  CalendarSelection sel[32];
  int nextmonth, nextyear;

  baseyear = closure;

  nextmonth = basemonth%12 + 1;
  nextyear = baseyear + ((basemonth == 12)?1:0);

  XtSetArg( args[0], XmNyear, baseyear );
  XtSetValues( cal1, args, 1 );

  sel[0].month = basemonth;
  sel[0].year = baseyear;
  sel[0].day = 0;
  getSelRecs( sel );
  CalSetSelInfo( cal1, SELECTED, sel );

  XtSetArg( args[0], XmNyear, nextyear );
  XtSetValues( cal2, args, 1 );

  sel[0].month = nextmonth;
  sel[0].year = nextyear;
  sel[0].day = 0;
  getSelRecs( sel );
  CalSetSelInfo( cal2, SELECTED, sel );
}

focus( i )
int i;
{
  XtSetKeyboardFocus( mainShell, fwidget[i] );
}

void focuscb( widget, closure, str )
Widget widget;
int closure;
XmAnyCallbackStruct *str;
{
/*
  char *key;

  key = '\t';
  if ( str->event->type == KeyPress )
    XLookupString( str->event, &key, 1, NULL, NULL );
*/

  if ( str->event->xkey.state & ShiftMask )
    focus( (closure+2)%3 ); /* go back one */
  else
    focus( (closure+1)%3 );
}


void togpref( widget, closure, call_data )
Widget widget;
int closure;
caddr_t call_data;
{
  int sa = 1;
  static int sy = 0, sm = 1, sg = 0, sd = 1;
  
  switch( closure )
    {
    case 1:
      sy = 1 - sy;
      XtSetArg( args[0], XmNshowYear, sy );
      break;
    case 2:
      sm = 1 - sm;
      XtSetArg( args[0], XmNshowMonth, sm );
      break;
    case 3:
      sd = 1 - sd;
      XtSetArg( args[0], XmNshowDays, sd );
      break;
    case 4:
      sg = 1 - sg;
      XtSetArg( args[0], XmNdrawGrid, sg );
      break;
    case 5:
      assume_day = 1 - assume_day;
      sa = 0;
      break;
    case 6:
      ampm = 1 - ampm;
      sa = 0;
      break;
    }

  if ( sa )
    {
      XtSetValues( cal1, args, 1 ); /* use array of cals */
      XtSetValues( cal2, args, 1 );
    }
}

/*
 * Clear selections made in all calendars except widget
 */
clearCalSels( widget )
Widget widget;
{
  CalendarSelection sel[32]; /* make global */
  int c = 0;

  while ( displayedcals[c] != NULL )
    {
      if ( displayedcals[c] != (CalendarWidget) widget )
	{
	  sel[0].day = 0; /* this is ok for clearing */
	  CalSetSelInfo( displayedcals[c], SELECTED, sel );
	}
      c++;
    }
}

/*
 * this exists to maintain selection coherence across multiple calendars
 */
void click( widget, closure, data )
Widget widget;
caddr_t closure, data;
{
    clearCalSels( widget ); /* clear selections on all other calendars */
    /* don't clear our concept of what is selected yet - we have to write
       the record back first; since we also write to the file at the same
       time, it's possible that we'll get an error and writerec will pop
       up a dialog box. If this happens, we won't get a mouseup and
       dateSelect won't get called, and things will get all screwed up.
       thus, writerec() and clearSelRecs() are called in dateselect and
       not here.
       this results in inconsistent state for a while, but there's not
       much that can be done about it. darn modal shit. */
}

void timeToStrC( t, str )
time t;
char *str;
{
  char ap[] = {' ','\0'};
  int h;

  if ( !ampm )
    {
      if ( t.minute == 0 )
	sprintf( str, "%d",
		 h);
      else
	sprintf( str, "%d:%02d",
		 t.hour, t.minute );
    }
  else
    {
      h = t.hour;
      if ( h < 12 )
	{
	  if ( h == 0 )
	    h = 12;
	  ap[0] = 'a';
	}
      else
	{
	  ap[0] = 'p';
	  if ( h > 12 )
	    h -= 12;
	}
      if ( t.minute == 0 )
	sprintf( str, "%d%s",
		 h, ap );
      else
	sprintf( str, "%d:%02d%s",
		 h, t.minute, ap );
    }
}

void printproc( widget, closure, data )
Widget widget;
caddr_t closure, data;
{
  static char *mns[12] = { "jan", "feb", "mar", "apr", "may", "jun",
			   "jul", "aug", "sep", "oct", "nov", "dec" };
  static char pname[50] = "nil";
  char call[80];
  static simpleEvent *simp = NULL, *ptr;
  FILE *f;
  char ststr[20], endstr[20];
  int err;
  int i, w;
  int month, year;

  XtSetArg( args[0], XmNmonth, &month );
  XtSetArg( args[1], XmNyear, &year );
  XtGetValues( displayedcals[0], args, 2 );

  sprintf( call, "Printing calendar for %d/%d.\nEnter printer name:",
	   month, year );
  if ( MuGetString( call, pname, 50, NULL ) )
    {
      f = fopen( "/tmp/tc.conv", "w" );

      fprintf( f, "year %d\n", year );

      for ( i = 1; i < 32; i++ )
	{
	  if ( simp != NULL )
	    freeSimple( simp );

	  simp = cvtEventToSimple( events, year,
				           month, i );

	  ptr = simp;
	  while ( ptr != NULL )
	    {
	      if ( ptr->info->desc != NULL )
		{
		  w = index( ptr->info->desc, '\n' ) - ptr->info->desc;
		  if ( w < 0 )
		    w = strlen( ptr->info->desc ); /* 500+ dies; bug */

		  if ( ptr->info->daylike )
		    fprintf( f, "%s %d %.*s\n",
			     mns[month-1], i,
			     w, ptr->info->desc );
		  else
		    {
		      timeToStrC( ptr->starttime, ststr );
		      timeToStrC( ptr->endtime, endstr );
		      fprintf( f, "%s %d %s-%s %.*s\n",
			       mns[month-1], i,
			       ststr, endstr,
			       w, ptr->info->desc );
		    }
		  ptr = ptr->next;
		}
	    }
	}
      fclose( f );

      sprintf( call, "pcal -f /tmp/tc.conv -m %d -y %d | lpr -P%s",
	       month, year, pname );
      if ( 0!= (err = system( call ) ) )
	{
	  sprintf( call, "Print attempt returned error: %d", err );
	  MyError( 0, call );
	}
    }
  if ( simp != NULL )
    {
      freeSimple( simp );
      simp = NULL; /* redundant? */
    }
}

void saveproc( widget, closure, data )
Widget widget;
caddr_t closure, data;
{
  if ( editingold )
    if ( W_BADENTRY != writerec() )
      {
	XmTextSetString( descr, NULL );
	XmTextSetString( stime, NULL );
	XmTextSetString( etime, NULL );
	focus( 0 );

	setuptoday();
      }
}

void newproc( widget, closure, data )
Widget widget;
caddr_t closure, data;
{
  if ( editingold )
    {
      if ( W_BADENTRY != writerec() )
	{
	  XmTextSetString( descr, NULL );
	  XmTextSetString( stime, NULL );
	  XmTextSetString( etime, NULL );

	  setuptoday();
	}
    }
  else
    {
      XmTextSetString( descr, NULL );
      XmTextSetString( stime, NULL );
      XmTextSetString( etime, NULL );
    }
  focus( 0 );
  editingold = False;
}

/*
 * sets up the list widget to contain today (displayed var)
 * sorting may want to go here
 */
setuptoday()
{
  simpleEvent *ptr;
  char *str;
  char ststr[20], endstr[20];
  char temp[500];
  int w;
  int i, j, pos, last;

  if ( displayed.day == 0 )
    sprintf( temp, "No day selected." );
  else
    sprintf( temp, "%s %d%s, %d", month( displayed.month ),
	     displayed.day,
	     suffix( displayed.day ),
	     displayed.year );

  str = XmStringCreateLtoR( temp,
			    XmSTRING_DEFAULT_CHARSET );
  XtSetArg( args[0], XmNlabelString, str );
  XtSetValues( header, args, 1 );

  XtSetArg( args[0], XmNselectedItems, NULL );
  XtSetArg( args[1], XmNselectedItemCount, 0 );
  XtSetValues( dayList, args, 2 );

  if ( numitems != 0 )
    {
      XtSetArg( args[0], XmNitems, NULL );
      XtSetArg( args[1], XmNitemCount, 0 );
      XtSetValues( dayList, args, 2 );
      numitems = 0;

      freeSimple( simp );
    }

  if ( displayed.day != 0 )
    {
      simp = cvtEventToSimple( events, displayed.year, displayed.month,
			       displayed.day );

      ptr = simp;
      while ( ptr != NULL )
	{
	  if ( ptr->info->desc != NULL )
	    {
	      w = index( ptr->info->desc, '\n' ) - ptr->info->desc;
	      if ( w < 0 )
		w = strlen( ptr->info->desc ); /* 500+ dies; bug */

	      if ( ptr->info->daylike )
		sprintf( temp, "          All day...          %.*s",
			 w, ptr->info->desc );
	      else
		{
		  timeToStr( ptr->starttime, ststr );
		  timeToStr( ptr->endtime, endstr );
		  sprintf( temp, "%s - %s   %.*s",
			   ststr, endstr,
			   w, ptr->info->desc );
		}

	      str = XmStringCreateLtoR( temp,
				        XmSTRING_DEFAULT_CHARSET );

	      itemarray[numitems++] = ptr;

	      XmListAddItem( dayList, str, 0 );
	      XtFree( str );
	    }
	  ptr = ptr->next;
	}
    }
}

int blank( str )
char *str;
{
  char *ptr;

  ptr = str;

  while ( isspace( *(ptr++) ) );
  if ( *(--ptr) == '\0' )
    return 1;
  else
    return 0;
}

int writerec()
{
  char *str;
  time s, e;
  desc d;

  if ( noselections() )
    {
      MuWarning( "There are no dates selected for this entry,\nso it can't be added yet." );
      return W_BADENTRY;
    }

  str = XmTextGetString( stime );
  if ( blank( str ) )
    {
      d.daylike = 1;
      s.hour = s.minute = s.second = 0;
      e.hour = e.minute = e.second = 0;
    }
  else
    {
      d.daylike = 0;
      strToTime( str, &s );
      XtFree( str );

      str = XmTextGetString( etime );
      strToTime( str, &e );
    }
  XtFree( str );

  str = XmTextGetString( descr );

  d.desc = str;
  d.scheduled = 1;

  d.activity = getAct();

  if ( *str != '\0' )
      addToEvents( &events,
		  selections, nummonths,
		  s, e,
		  &d );
  XtFree( str );

  if ( editingold )
    deleteFromEvents( &events, editingwhat );

  if ( dumpEventsToFile( events, filename ) )
    return W_SUCCESS;
  else
    {
      MyError( 0, "The calendar file was not written due to an error.\nUse Save to try again." );
      return W_FILE;
    }
}

void showEdit( widget, closure, data )
Widget widget;
caddr_t closure;
caddr_t data;
{
  XtManageChild( parmForm );
}

void hideEdit( widget, closure, data )
Widget widget;
caddr_t closure;
caddr_t data;
{
  XtUnmanageChild( parmForm );
}

void deleteEnt( widget, closure, data )
Widget widget;
caddr_t closure;
caddr_t data;
{
  if ( editingold )
    {
      deleteFromEvents( &events, editingwhat );

      XmTextSetString( descr, NULL );
      XmTextSetString( stime, NULL );
      XmTextSetString( etime, NULL );
      focus( 0 );

      setuptoday();
      editingold = False;

      if ( dumpEventsToFile( events, filename ) )
	return (void)W_SUCCESS;
      else
	{
	  MyError( 0, "The calendar file was not written due to an error.\nUse Save to try again." );
	  return (void)W_FILE;
	}
    }
}

void dateExtSelect( widget, closure, data )
Widget widget;
caddr_t closure;
caddr_t data;
{
  CalendarSelection sel[32];

  CalGetSelInfo( (CalendarWidget) widget, SELECTED, sel );
  setSelRecs( sel );
  /* printSelRecs(); */
}

void dateSelect( widget, closure, data )
Widget widget;
caddr_t closure;
caddr_t data;
{
  CalendarSelection sel[32];

  if ( editingold )
    writerec(); /* needs to be conditional on changed */
                /* also has to be here to happen before selections cleared */
                /* execution here implies execution in dateSelect (which
		   turned out to be wrong...) */
  /* it may not have written - life goes on! */

  clearSelRecs(); /* will get reset for this widget in dateSelect (ha ha) */

  CalGetSelInfo( (CalendarWidget) widget, SELECTED, sel );
  setSelRecs( sel );
  /* printSelRecs(); */

/* displayed will really have to be initialized to today or something */
if ( displayed.month != sel[0].month ||
    displayed.day != sel[0].day ||
    displayed.year != sel[0].year ||
    editingold )
  {
    if ( editingold )
      {
	/* writerec(); done in click */
	XmTextSetString( descr, NULL );
	XmTextSetString( stime, NULL );
	XmTextSetString( etime, NULL );
	focus( 0 );
      }
    displayed = sel[0];
    setuptoday();
    editingold = False;
  }
}

void listsel( widget, closure, reason )
Widget widget;
XmListCallbackStruct *reason;
caddr_t closure;
{
  simpleEvent *s;
  char temp[20];
  int c = 0, month, year;
  CalendarSelection sel[32];

/*
  if ( editingold )
    writerec();
*** not implemented in this interface */

  s = itemarray[reason->item_position - 1];
  editingold = True;
  editingwhat = s->info;

  setAct( s->info->activity );

  if ( s->info->daylike )
    {
      XmTextSetString( stime, NULL );
      XmTextSetString( etime, NULL );
    }
  else
    {
      timeToStr( s->starttime, temp );
      XmTextSetString( stime, temp );

      timeToStr( s->endtime, temp );
      XmTextSetString( etime, temp );
    }

  XmTextSetString( descr, s->info->desc );

  getalldates( events, s->info, selections, &nummonths );
  while ( displayedcals[c] != NULL )
    {
      XtSetArg( args[0], XmNmonth, &month );
      XtSetArg( args[1], XmNyear, &year );
      XtGetValues( displayedcals[c], args, 2 );

      sel[0].month = month;
      sel[0].year = year;
      getSelRecs( sel );

      CalSetSelInfo( displayedcals[c], SELECTED, sel );      
      c++;
    }
  focus( 0 );
}

void addproc( widget, closure, call_data )
Widget widget;
caddr_t closure;
caddr_t call_data;
{
  if ( W_BADENTRY != writerec() )
    {
      XmTextSetString( descr, NULL );
      XmTextSetString( stime, NULL );
      XmTextSetString( etime, NULL );
      focus( 0 );
      setuptoday();
      editingold = False;
    }
}

main(argc,argv)
int argc;
char **argv;
{
  struct timeval time;
  struct tm *timestr;
  int i, n;
  char *str;
  char temp[20];
  CalendarSelection sel[32];

/*
 * File stuff
 */
  sprintf( filename, "%s/.tc", getenv( "HOME" ) ); /* errors... */

  events = NULL;
  addFileToEvents( filename, &events ); /* put error stuff here */

/*
 * Main mode window
 */

  mainShell = XtInitialize( argv[0], "tc", NULL, 0, &argc, argv );

  top[0] = mainShell;

  disp = XtDisplay( mainShell );
  screennum = DefaultScreen( disp );
  screen = ScreenOfDisplay( disp, screennum );
  root = RootWindow( disp, screennum );

  mainForm = XtCreateManagedWidget( "form",
			      xmFormWidgetClass,
			      mainShell, NULL, 0 );

  /* menu */
  MuRegisterMenuCallbacks( callbacks, XtNumber(callbacks) );
  menubar = MuCreateMenuBar( mainForm, "menubar", NULL, 0);
  XtManageChild(menubar);

  /* calendar */

  cal1 = XtCreateManagedWidget( "cal1", calendarWidgetClass,
			       mainForm, NULL, 0 );

  n = 0;
  XtSetArg( args[n], XmNmonth, &basemonth ); n++;
  XtSetArg( args[n], XmNyear, &baseyear ); n++;
  XtGetValues( cal1, args, n );

  n = 0;
  XtSetArg( args[n], XmNmonth, basemonth%12 + 1 ); n++;
  XtSetArg( args[n], XmNyear, baseyear + ((basemonth == 12)?1:0) ); n++;

  cal2 = XtCreateManagedWidget( "cal2", calendarWidgetClass,
			       mainForm, args, n );

  XtAddCallback( cal1, XmNdateProc, dateSelect, NULL );
  XtAddCallback( cal2, XmNdateProc, dateSelect, NULL );
  XtAddCallback( cal1, XmNdateExtProc, dateExtSelect, NULL );
  XtAddCallback( cal2, XmNdateExtProc, dateExtSelect, NULL );
  XtAddCallback( cal1, XmNclickProc, click, NULL );
  XtAddCallback( cal2, XmNclickProc, click, NULL );

  /* text editing widgets */

/* XmNvalueChangedCallback */
  n = 0;
  XtSetArg( args[n], XmNeditMode, XmMULTI_LINE_EDIT ); n++;

  descr = XtCreateManagedWidget( "description",
			         xmTextWidgetClass,
			         mainForm, args, 1 );

  XtAddCallback( descr, XmNactivateCallback, focuscb, 2 );

  str = XmStringCreateLtoR( "From",
			   XmSTRING_DEFAULT_CHARSET );
  n = 0;
  XtSetArg( args[n], XmNlabelString, str ); n++;
  from = XtCreateManagedWidget( "from",
			        xmLabelWidgetClass,
			        mainForm, args, n );
  XmStringFree( str );

  stime = XtCreateManagedWidget( "start",
			         xmTextWidgetClass,
			         mainForm, NULL, 0 );
  XtAddCallback( stime, XmNactivateCallback, focuscb, 0 );

  to = XtCreateManagedWidget( "to",
			      xmLabelWidgetClass,
			      mainForm, NULL, 0 );

  etime = XtCreateManagedWidget( "end",
			         xmTextWidgetClass,
			         mainForm, NULL, 0 );
  XtAddCallback( etime, XmNactivateCallback, focuscb, 1 );

  fwidget[0] = stime;
  fwidget[1] = etime;
  fwidget[2] = descr;

  /* add button widget */
  add = XtCreateManagedWidget( "enter",
			       xmPushButtonWidgetClass,
			       mainForm, NULL, 0 );

  XtAddCallback( add, XmNactivateCallback, addproc, NULL );

  /* new button widget */
  new = XtCreateManagedWidget( "new",
			       xmPushButtonWidgetClass,
			       mainForm, NULL, 0 );

  XtAddCallback( new, XmNactivateCallback, newproc, NULL );

  /* list widget header label */
  header = XtCreateManagedWidget( "today",
				 xmLabelWidgetClass,
				 mainForm, NULL, 0 );

  /* list widget */

  n = 0;
  XtSetArg( args[n], XmNlistSizePolicy, XmCONSTANT ); n++;
  XtSetArg( args[n], XmNselectionPolicy, XmSINGLE_SELECT ); n++;

/*
  dayList = XtCreateManagedWidget( "list",
			        xmListWidgetClass,
			        mainForm, args, n );
*/

  dayList = XmCreateScrolledList( mainForm, "list", args, n );
  XtManageChild( dayList );

  XtAddCallback( dayList, XmNsingleSelectionCallback, listsel, NULL );

/*
 * Entry parameters widgets
 */

  parmForm = XmCreateFormDialog( mainShell, "parmform", NULL, 0 );

/*
  schedToggle = XtCreateManagedWidget( "sched",
				       xmToggleButtonWidgetClass,
				       parmForm, NULL, 0 );
*/

  minHead = XtCreateManagedWidget( "minHead",
				   xmLabelWidgetClass,
				   parmForm, NULL, 0 );

  for ( i = 0; i < MAXMINS; i++ )
    {
      sprintf( temp, "min%d", mins[i] );
      minToggle[i] = XtCreateManagedWidget( temp,
					    xmToggleButtonWidgetClass,
					    parmForm, NULL, 0 );
    }

  createApptWidgets( mainShell );
/*
 * end widget creation
 */

  MuSetEmacsBindings( (Widget) stime );
  MuSetEmacsBindings( (Widget) etime );
  MuSetEmacsBindings( (Widget) descr );

  displayedcals[0] = (CalendarWidget) cal1;
  displayedcals[1] = (CalendarWidget) cal2;
  displayedcals[2] = NULL;

  nummonths = 0;

  for ( n = 0; n < MAXMONTHS; n++ )
    selections[n] = NULL;

  XtInstallAccelerators( descr, add );
  XtInstallAccelerators( stime, add );
  XtInstallAccelerators( etime, add );
  MuInitialize( mainShell );

  XtRealizeWidget( mainShell );

  focus( 0 );

  gettimeofday( &time, NULL );
  timestr = localtime( &time.tv_sec );
  displayed.month = timestr->tm_mon + 1;
  displayed.year = timestr->tm_year + 1900;
  displayed.day = timestr->tm_mday;
  sel[0] = displayed;
  sel[1].day = 0;
  CalSetSelInfo( (CalendarWidget) cal1, SELECTED, sel );
  setSelRecs( sel );
  setuptoday();

  XtMainLoop();
}
